use crate::models::events::Event;
use std::{
    collections::{HashMap, HashSet},
    future::Future,
};
use tracing::info;

type StateMap<T> = HashMap<StateTuple, T>;

#[derive(Debug, PartialEq, Eq, Hash, Clone)]
pub struct StateTuple {
    event_type: String,
    state_key: String,
}

#[derive(Debug, Clone, PartialEq, Eq, Hash)]
pub struct EventId {
    id: Box<str>,
}

#[tracing::instrument(skip(state_sets, auth_chain_sets, get_event_callback))]
pub async fn resolve<F, Fut>(
    room_id: &str, // TODO: own type
    state_sets: Vec<StateMap<EventId>>,
    auth_chain_sets: Vec<HashSet<EventId>>,
    get_event_callback: F,
) -> StateMap<EventId>
where
    F: Fn(&EventId) -> Fut,
    Fut: Future<Output = Option<Event>>,
{
    info!("Calculating conflicted state");

    let (unconflicted_state, conflicted_state) = separate_state(&state_sets);

    if conflicted_state.is_empty() {
        return unconflicted_state;
    }

    info!("{} conflicted_state entries", conflicted_state.len());
    info!("Calculating auth_chain differences");

    let conflicted_set =
        get_auth_chain_differences(auth_chain_sets).chain(conflicted_state.into_values().flatten());
    let mut conflicted = HashSet::new();
    for eid in conflicted_set {
        if let Some(event) = get_event_callback(&eid).await {
            conflicted.insert(event);
        }
    }

    todo!()
}

/// separates states from multiple state_maps into unconflicted and conflicted state
///
/// For the set of all state_tuples find all event_ids.
/// If one event_id is found it is unconflicted, otherwise it is conflicted
fn separate_state(
    state_sets: &[StateMap<EventId>],
) -> (StateMap<EventId>, StateMap<HashSet<EventId>>) {
    let mut unconflicted_state: StateMap<EventId> = StateMap::new();
    let mut conflicted_state: HashMap<StateTuple, HashSet<EventId>> = StateMap::new();

    for key in state_sets
        .iter()
        .flat_map(HashMap::keys)
        .map(ToOwned::to_owned)
        .collect::<HashSet<StateTuple>>()
    {
        let mut event_ids: HashSet<EventId> = state_sets
            .iter()
            .filter_map(|state_set| state_set.get(&key))
            .map(ToOwned::to_owned)
            .collect();

        if event_ids.len() == 1 {
            unconflicted_state.insert(key, event_ids.into_iter().next().expect("len() is 1"));
        } else {
            conflicted_state.insert(key, event_ids);
        }
    }

    (unconflicted_state, conflicted_state)
}

fn get_auth_chain_differences(
    auth_chain_sets: Vec<HashSet<EventId>>,
) -> impl Iterator<Item = EventId> {
    let num_sets = auth_chain_sets.len();

    let mut id_counts: HashMap<EventId, usize> = HashMap::new();
    for id in auth_chain_sets.into_iter().flatten() {
        *id_counts.entry(id).or_default() += 1;
    }

    id_counts
        .into_iter()
        .filter_map(move |(id, count)| (count < num_sets).then(move || id))
}

fn is_control_event() {}
